#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `mlconjug` package."""

import pytest
import os

from click.testing import CliRunner

from collections import OrderedDict

from mlconjug import mlconjug, PyVerbiste
from mlconjug import cli


conjug_aller = OrderedDict([
    ('infinitive', OrderedDict([('infinitive-present', 'aller')])),
    ('indicative', OrderedDict([
        ('present', OrderedDict([('1s', 'vais'), ('2s', 'vas'), ('3s', 'va'),
                                 ('1p', 'allons'), ('2p', 'allez'), ('3p', 'vont')])),
        ('imperfect', OrderedDict([('1s', 'allais'), ('2s', 'allais'), ('3s', 'allait'),
                                   ('1p', 'allions'), ('2p', 'alliez'), ('3p', 'allaient')])),
        ('future', OrderedDict([('1s', 'irai'), ('2s', 'iras'), ('3s', 'ira'),
                                ('1p', 'irons'), ('2p', 'irez'), ('3p', 'iront')])),
        ('simple-past', OrderedDict([('1s', 'allai'), ('2s', 'allas'), ('3s', 'alla'),
                                     ('1p', 'allâmes'), ('2p', 'allâtes'), ('3p', 'allèrent')]))])),
    ('conditional', OrderedDict([
        ('present', OrderedDict([('1s', 'irais'), ('2s', 'irais'), ('3s', 'irait'),
                                 ('1p', 'irions'), ('2p', 'iriez'), ('3p', 'iraient')]))])),
    ('subjunctive', OrderedDict([
        ('present', OrderedDict([('1s', 'aille'), ('2s', 'ailles'), ('3s', 'aille'),
                                 ('1p', 'allions'), ('2p', 'alliez'), ('3p', 'aillent')])),
        ('imperfect', OrderedDict([('1s', 'allasse'), ('2s', 'allasses'), ('3s', 'allât'),
                                   ('1p', 'allassions'), ('2p', 'allassiez'), ('3p', 'allassent')]))])),
    ('imperative', OrderedDict([
        ('imperative-present', OrderedDict([('2s :', 'va'), ('1p :', 'allons'), ('2p :', 'allez')]))])),
    ('participle', OrderedDict([
        ('present-participle', 'allant'),
        ('past-participle', OrderedDict([('ms :', 'allé'), ('mp :', 'allés'),
                                         ('fs :', 'allée'), ('fp :', 'allées')]))]))])

class TestPyVerbiste:
    verbiste = PyVerbiste.Verbiste(language='fr')
    def test_init_verbiste(self):
        assert len(self.verbiste.templates) == len(self.verbiste.conjugations) == 149
        assert self.verbiste.templates[0] == ':aller'
        assert self.verbiste.templates[-1] == 'écri:re'
        assert self.verbiste.conjugations[':aller'] == conjug_aller
        assert len(self.verbiste.verbs) == 7015
        assert self.verbiste.verbs['abaisser'] == {'template': 'aim:er', 'root': 'abaiss'}

    def test_get_verb_info(self):
        verb_info = self.verbiste.get_verb_info('aller')
        assert verb_info == PyVerbiste.VerbInfo('aller', '', ':aller')
        assert self.verbiste.get_verb_info('cacater') is None

    def test_get_conjug_info(self):
        conjug_info = self.verbiste.get_conjug_info(':aller')
        assert conjug_info == self.verbiste.conjugations[':aller']
        assert self.verbiste.get_conjug_info(':cacater') is None

class TestVerb:
    verbiste = PyVerbiste.Verbiste(language='fr')
    def test_verbinfo(self):
        test_verb_info = self.verbiste.get_verb_info('aller')
        test_conjug_info = self.verbiste.get_conjug_info(':aller')
        test_verb = PyVerbiste.Verb(test_verb_info, test_conjug_info)
        assert test_verb.conjug_info == conjug_aller


class TestEndingCountVectorizer:
    ngrange = (2, 7)
    vectorizer = mlconjug.EndingCountVectorizer(analyzer="char", binary=True, ngram_range=ngrange)
    def test_char_ngrams(self):
        ngrams = self.vectorizer._char_ngrams('aller')
        assert 'ller' in ngrams


class TestConjugator:
    conjugator = mlconjug.Conjugator()
    def test_conjugate(self):
        test_verb = self.conjugator.conjugate('aller')
        assert isinstance(test_verb, PyVerbiste.Verb)
        assert test_verb.conjug_info == conjug_aller
        assert test_verb.verb_info == PyVerbiste.VerbInfo('aller', '', ':aller')

    def test_set_model(self):
        self.conjugator.set_model(mlconjug.Model())
        assert isinstance(self.conjugator.model, mlconjug.Model)


class TestDataSet:
    data_set = mlconjug.DataSet(mlconjug.Verbiste())
    def test_construct_dict_conjug(self):
        self.data_set.construct_dict_conjug()
        assert 'aller' in self.data_set.dict_conjug[':aller']

    def test_split_data(self):
        self.data_set.split_data()
        assert self.data_set.test_input is not None
        assert self.data_set.train_input is not None
        assert self.data_set.test_labels is not None
        assert self.data_set.train_labels is not None


class TestModel:
    vectorizer = mlconjug.EndingCountVectorizer(analyzer="char", binary=True,
                                                ngram_range=(2, 7))
    # Feature reduction
    feature_reductor = mlconjug.SelectFromModel(
        mlconjug.LinearSVC(penalty="l1", max_iter=3000, dual=False, verbose=2))
    # Prediction Classifier
    classifier = mlconjug.SGDClassifier(loss="log", penalty='elasticnet',
                                        alpha=1e-5, random_state=42)
    # Initialize Model
    model = mlconjug.Model(vectorizer, feature_reductor, classifier)
    dataset = mlconjug.DataSet(mlconjug.Verbiste())
    dataset.construct_dict_conjug()
    dataset.split_data(proportion=0.9)

    def test_train(self):
        self.model.train(self.dataset.test_input, self.dataset.test_labels)
        assert isinstance(self.model, mlconjug.Model)

    def test_predict(self):
        result = self.model.predict(['aimer',])
        assert self.dataset.templates[result[0]] == 'aim:er'



def test_command_line_interface():
    """Test the CLI."""
    verb = 'aller'
    runner = CliRunner()
    result = runner.invoke(cli.main, [verb])
    assert result.exit_code == 0
    assert 'allassions' in result.output
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert 'Console script for mlconjug.' in help_result.output



# -*- coding: utf-8 -*-

"""Top-level package for MLConjug."""

__author__ = """SekouD"""
__email__ = 'sekoud.python@gmail.com'
__version__ = '1.2.1'

from .mlconjug import EndingCountVectorizer, DataSet, Model, LinearSVC, SGDClassifier, SelectFromModel, precision_recall_fscore_support, Conjugator
from .PyVerbiste import Verbiste



# -*- coding: utf-8 -*-

"""
PyVerbiste.
A Python library for conjugating verbs.
It uses verbiste xml conjugation data.
Verbiste can be downloaded at https://perso.b2b2c.ca/~sarrazip/dev/verbiste.html

"""

import copy
import xml.etree.ElementTree as ET
import codecs
from collections import OrderedDict
import pkg_resources
from _io import BufferedReader

resource_package = __name__
verbs_resource_path = {'fr': '/'.join(('data', 'verbiste', 'verbs-fr.xml')),
                       'it': '/'.join(('data', 'verbiste', 'verbs-it.xml'))}
conjugations_resource_path = {'fr': '/'.join(('data', 'verbiste', 'conjugation-fr.xml')),
                              'it': '/'.join(('data', 'verbiste', 'conjugation-it.xml'))}


class Verbiste:
    """
    This is the class handling the Verbiste xml files.

    :param language: string.
        The language of the conjugator. The default value is fr for French.
    :param subject: string.
        Value can be either 'default' or 'pronoun' depending on if you want to have full pronouns or abbreviated pronouns.

    """

    def __init__(self, language='fr', subject='default'):
        self.subject = subject
        self.verbs = {}
        self.conjugations = OrderedDict()
        verbs_file = pkg_resources.resource_stream(
        resource_package, verbs_resource_path[language])
        self._load_verbs(verbs_file)
        conjugations_file = pkg_resources.resource_stream(
        resource_package, conjugations_resource_path[language])
        self._load_conjugations(conjugations_file)
        self.templates = sorted(self.conjugations.keys())
        self.model = None

    def _load_verbs(self, verbs_file):
        """
        Load and parses the verbs from xml file.

        :param verbs_path: string or path object.
            Path to the verbs xml file.
        """
        verbs_dic = {}
        if isinstance(verbs_file, BufferedReader):
            verbs_dic = self._parse_verbs(verbs_file)
        else:
            with codecs.open(verbs_file, "r", encoding='utf-8') as file:
                verbs_dic = self._parse_verbs(file)
        self.verbs = verbs_dic

    def _parse_verbs(self, file):
        """
        Parses XML file

        :param file: XML file containing the verbs.
        :return: OrderedDict.
        """
        verbs_dic = {}
        xml = ET.parse(file)
        for verb in xml.findall("v"):
            verb_name = verb.find("i").text
            template = verb.find("t").text
            index = - len(template[template.index(":") + 1:])
            root = verb_name[:index]
            verbs_dic[verb_name] = {"template": template, "root": root}
        return verbs_dic


    def _load_conjugations(self, conjugations_file):
        """
        Load and parses the conjugations from xml file.

        :param conjugations_path: string or path object.
            Path to the conjugation xml file.
        """
        if isinstance(conjugations_file, BufferedReader):
            conjugations_dic = self._parse_conjugations(conjugations_file)
        else:
            with codecs.open(conjugations_file, "r", encoding='utf-8') as file:
                conjugations_dic = self._parse_conjugations(file)
        self.conjugations = conjugations_dic

    def _parse_conjugations(self, file):
        """
        Parses XML file

        :param file: XML file containing the conjugation templates
        :return: OrderedDict
        """
        conjugations_dic = {}
        xml = ET.parse(file)
        for template in xml.findall("template"):
            template_name = template.get("name")
            conjugations_dic[template_name] = OrderedDict()
            for mood in list(template):
                conjugations_dic[template_name][mood.tag] = OrderedDict()
                for tense in list(mood):
                    conjugations_dic[template_name][mood.tag][tense.tag] = self._load_tense(mood, tense)
        return conjugations_dic


    def _load_tense(self, mood, tense):
        """
        Parses XML objects and extract mood and tense information to create a dict containing all inflected forms of the verb.

        :param mood: xml.etree.ElementTree Tag object.
        :param tense: xml.etree.ElementTree Tag object.
        :return: OrderedDict or None.
        """
        # TODO: add pronouns for languages other than french.
        persons = list(tense)
        if tense.tag in ('infinitive-present', 'present-participle'):
            if persons[0].find("i") is not None:
                conjug = persons[0].find("i").text
            else:
                conjug = None
        elif tense.tag == 'imperative-present':
            pronouns = ["2s :", "1p :", "2p :"]
            conjug = OrderedDict((pers, term.find("i").text if term.find("i") is
                                                               not None else None) for pers, term in
                                 zip(pronouns, persons))
        elif tense.tag == 'past-participle':
            pronouns = ["ms :", "mp :", "fs :", "fp :"]
            conjug = OrderedDict((pers, term.find("i").text if term.find("i") is
                                                               not None else None) for pers, term in
                                 zip(pronouns, persons))
        elif len(persons) == 6:
            if self.subject == 'pronoun':
                pronouns = ["je", "tu", "il (elle, on)", "nous", "vous", "ils (elles)"]
                if mood.tag == 'subjunctive':
                    prefix1 = "que "
                    prefix2 = "qu' "
                    exceptions = ("il (elle, on)", "ils (elles)")
                    pronouns = [prefix1 + elmt if elmt not in exceptions else prefix2 + elmt for elmt in pronouns]
            else:
                pronouns = ["1s", "2s", "3s", "1p", "2p", "3p"]
            conjug = OrderedDict((pers, term.find("i").text if term.find("i") is
                                                               not None else None) for pers, term in
                                 zip(pronouns, persons))
        return conjug

    def get_verb_info(self, verb):
        """
        Gets verb information and returns a VerbInfo instance.

        :param verb: string.
            Verb to conjugate.
        :return: VerbInfo object or None.
        """
        if verb not in self.verbs.keys():
            return None
        else:
            infinitive = verb
            root = self.verbs[verb]['root']
            template = self.verbs[verb]['template']
            verb_info = VerbInfo(infinitive, root, template)
            return verb_info

    def get_conjug_info(self, template):
        """
        Gets conjugation information corresponding to the given template.

        :param template: string.
            Name of the verb ending pattern.
        :return: OrderedDict or None.
        """
        if template not in self.conjugations.keys():
            return None
        else:
            info = copy.deepcopy(self.conjugations[template])
            return info


class VerbInfo(object):
    """
    This class defines the Verbiste verb information structure.

    :param infinitive: string.
        Infinitive form of the verb.
    :param root: string.
        Lexical root of the verb.
    :param template: string.
        Name of the verb ending pattern.

    """
    __slots__ = ('infinitive', 'root', 'template')

    def __init__(self, infinitive, root, template):
        self.infinitive = infinitive
        self.root = root
        self.template = template

    def __eq__(self, other):
        if self.infinitive == other.infinitive \
            and self.root == other.root \
            and self.root == other.root:
            return True
        else:
            return False


class Verb(object):
    """
    This class defines the Verb Object.

    :param verb_info: VerbInfo Object.
    :param conjug_info: OrderedDict.

    """
    __slots__ = ('name', 'verb_info', 'conjug_info')

    def __init__(self, verb_info, conjug_info):
        self.name = verb_info.infinitive
        self.verb_info = verb_info
        self.conjug_info = conjug_info
        self._load_conjug()

    def _load_conjug(self):
        """
        Populates the inflected forms of the verb.

        """
        for mood, tense in self.conjug_info.items():
            for tense, persons in tense.items():
                if isinstance(persons, dict):
                    for pers, term in persons.items():
                        if term is not None:
                            persons[pers] = self.verb_info.root + term
                        else:
                            persons[pers] = None
                elif isinstance(persons, str):
                    self.conjug_info[mood][tense] = self.verb_info.root + persons
                    pass


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

"""Console script for mlconjug."""

import click
import mlconjug
from pprint import pprint


@click.command()
@click.argument('verb')
@click.option('-l', '--language',
              default='fr',
              help='The language for the conjugation model.',
              type=click.STRING)
def main(verb, language):
    """Console script for mlconjug."""
    conjugator = mlconjug.Conjugator(language)
    result = conjugator.conjugate(verb)
    pprint(result.conjug_info)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Click>=6.0',
    'numpy==1.14.3',
    'scipy==1.1.0',
    'scikit-learn==0.19.',
]

setup_requirements = [
    'pytest-runner',
    # TODO(SekouD): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    'pytest',
    'Sphinx',
    'docutils',
    'pytest',
    'pytest-cov',
    'Click>=6.0',
    'lxml',
]

setup(
    name='mlconjug',
    version='1.2.1',
    description="A Python library to conjugate French (and many other Romance languages) verbs using Machine Learning techniques.",
    long_description=readme + '\n\n' + history,
    author="SekouD",
    author_email='sekoud.python@gmail.com',
    url='https://github.com/SekouD/mlconjug',
    packages=find_packages(include=['mlconjug']),
    entry_points={
        'console_scripts': [
            'mlconjug=mlconjug.cli:main'
        ]
    },
    package_data={'verbiste_data': ['mlconjug/data/verbiste/*'],
                  'trained_models': 'mlconjug/data/models/*'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='mlconjug conjugation conjugaison verbs verbes ML machine learning NLP verbiste sklearn',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Education',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Utilities',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Natural Language :: French',
        'Natural Language :: Spanish',
        'Natural Language :: Italian',
        'Natural Language :: Portuguese',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



# -*- coding: utf-8 -*-

"""Main module."""

from .PyVerbiste import Verbiste, Verb, VerbInfo

from sklearn.feature_selection import SelectFromModel
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.svm import LinearSVC
from sklearn.linear_model import SGDClassifier
from sklearn.pipeline import Pipeline
from sklearn.metrics import precision_recall_fscore_support

import random
from collections import defaultdict
import pickle

import pkg_resources

resource_package = __name__
pre_trained_model_path = {'fr': '/'.join(('data', 'models', 'trained_model-fr.pickle')),
                          'it': '/'.join(('data', 'models', 'trained_model-it.pickle'))}


class Conjugator:
    """
    This is the main class of the project.
    The class manages the Verbiste data set and provides an interface with the scikit-learn model.
    If no parameters are provided

    :param language:
        Language of the conjugator. The default language is 'fr' for french.
    :param model:
        A user provided model if the user has trained his own model.

    """
    def __init__(self, language='fr', model=None):
        self.verbiste = Verbiste(language=language)
        self.data_set = DataSet(self.verbiste)
        self.data_set.construct_dict_conjug()
        self.data_set.split_data(proportion=0.9)
        if not model:
            model = pickle.loads(pkg_resources.resource_stream(
                resource_package, pre_trained_model_path[language]).read())
        self.model = model

    def conjugate(self, verb):
        """
        This is the main method of this class.
        It first checks to see if the verb is in Verbiste. If it is not, and a pre-trained scikit-learn model has been supplied,
        the method then calls the model to predict the conjugation class of the provided verb.
        Returns a Verb object.

        :param verb: string.
            Verb to conjugate.
        :return verb_object: Verb object or None.
        """
        infinitive = None
        if verb not in self.verbiste.verbs.keys():
            if self.model is None:
                return None
            predicted = self.model.predict([verb])[0]
            template = self.verbiste.templates[predicted]
            index = - len(template[template.index(":") + 1:])
            root = verb[:index]
            verb_info = VerbInfo(verb, root, template)
            conjug_info = self.verbiste.get_conjug_info(verb_info.template)
        else:
            infinitive = verb
            verb_info = self.verbiste.get_verb_info(infinitive)
            if verb_info is None:
                return None
            conjug_info = self.verbiste.get_conjug_info(verb_info.template)
            if conjug_info is None:
                return None
        verb_object = Verb(verb_info, conjug_info)
        return verb_object

    def set_model(self, model):
        """
        Assigns the provided pre-trained scikit-learn model to be able to conjugate unknown verbs.

        :param model: scikit-learn Classifier or Pipeline.
        """
        self.model = model
        return



class EndingCountVectorizer(CountVectorizer):
    """
    Custom Vectorizer optimized for extracting verbs features.
    The Vectorizer subclasses sklearn.feature_extraction.text.CountVectorizer
    As in Romance languages verbs are inflected by adding a morphological suffix,
    the vectorizer extracts verb endings and produces a vector representation of the verb with binary features.
    The features are the verb ending ngrams. (ngram_range is set at class initialization).

    """
    def _char_ngrams(self, verb):
        """
        Parses a verb and returns the ending ngrams.

        :param verb: string.
            Verb to vectorize.
        :return: list.
            Final ngrams of the verb.
        """
        verb = self._white_spaces.sub(" ", verb)
        verb_len = len(verb)
        ngrams = []
        min_n, max_n = self.ngram_range
        for n in range(min_n, min(max_n + 1, verb_len + 1)):
            ngram = verb[-n:]
            ngrams.append(ngram)
        return ngrams



class DataSet:
    """
    This class holds and manages the data set.
    Defines helper functions for managing Machine Learning Tasks.

    :param VerbisteObj:

    """
    def __init__(self, VerbisteObj):
        self.verbiste = VerbisteObj
        self.verbes = self.verbiste.verbs.keys()
        self.templates = sorted(self.verbiste.conjugations.keys())
        self.liste_verbes = []
        self.liste_templates = []
        self.dict_conjug = []
        self.train_input = []
        self.train_labels = []
        self.test_input = []
        self.test_labels = []

    def construct_dict_conjug(self):
        """
        Populates the dictionary containing the conjugation templates.

        """
        conjug = defaultdict(list)
        for verbe, info_verbe in self.verbiste.verbs.items():
            self.liste_verbes.append(verbe)
            self.liste_templates.append(
                self.templates.index(info_verbe["template"]))
            conjug[info_verbe["template"]].append(verbe)
        self.dict_conjug = conjug
        return

    def split_data(self, threshold=8, proportion=0.5):
        """
        Splits the data into a training and a testing set.

        :param threshold: int.
            Minimum size of conjugation class to be split.
        :param proportion: float.
            Proportion of samples in the training set. Must be between 0 and 1.
        """
        if proportion <= 0 or proportion >= 1:
            raise ValueError('The split proportion must be between 0 and 1')
        self.min_threshold = threshold
        self.split_proportion = proportion
        train_set = []
        test_set = []
        for template, lverbes in self.dict_conjug.items():
            if len(lverbes) <= threshold:
                for verbe in lverbes:
                    train_set.append((verbe, template))
            else:
                index = round(len(lverbes) * proportion)
                for verbe in lverbes[:index]:
                    train_set.append((verbe, template))
                for verbe in lverbes[index:]:
                    test_set.append((verbe, template))
        random.shuffle(train_set)
        random.shuffle(test_set)
        self.train_input = [elmt[0] for elmt in train_set]
        self.train_labels = [self.templates.index(elmt[1]) for elmt in
                             train_set]
        self.test_input = [elmt[0] for elmt in test_set]
        self.test_labels = [self.templates.index(elmt[1]) for elmt in test_set]
        return


class Model(object):
    """
    This class manages the scikit-learn model.
    The Pipeline includes a feature vectorizer, a feature selector and a classifier.
    If any of the vectorizer, feature selector or classifier is not supplied at instance declaration,
    the __init__ method will provide good default value that gets more than 98% prediction accuracy.

    :param vectorizer: scikit-learn Vectorizer.
    :param feature_selector: scikit-learn Classifier with a fit_transform() method
    :param classifier: scikit-learn Classifier with a predict() method
    """
    def __init__(self, vectorizer=None, feature_selector=None, classifier=None):
        if not vectorizer:
            vectorizer = EndingCountVectorizer(analyzer="char", binary=True, ngram_range=(2, 7))
        if not feature_selector:
            feature_selector = SelectFromModel(LinearSVC(penalty='l1', max_iter=12000, dual=False, verbose=2))
        if not classifier:
            classifier = SGDClassifier(loss='log', penalty='elasticnet', l1_ratio=0.15, max_iter=4000, alpha=1e-5, random_state=42, verbose=2)
        self.model = Pipeline([('vectorizer', vectorizer),
                               ('feature_selector', feature_selector),
                               ('classifier', classifier)])

    def train(self, samples, labels):
        """
        Trains the model on the supplied samples and labels.

        :param samples: list.
            List of verbs.
        :param labels: list.
            List of verb templates.
        """
        self.model = self.model.fit(samples, labels)
        return

    def predict(self, verbs):
        """
        Predicts the conjugation class of the provided list of verbs.

        :param verbs: list.
            List of verbs.
        :return: list.
            List of predicted conjugation groups.
        """
        prediction = self.model.predict(verbs)
        return prediction


if __name__ == "__main__":
    pass
